/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// sbar.c -- status bar code

#include "globaldef.h"

int		sb_scaled;
int			sb_updates;		// if >= vid.numpages, no update needed

#define STAT_MINUS		10	// num frame for '-' stats digit
qpic_t	*sb_nums[2][11];

qpic_t	*sb_colon, *sb_slash;
qpic_t	*sb_ibar;
qpic_t	*sb_sbar;
qpic_t	*sb_scorebar;


qpic_t	*sb_weapons[7][8];   // 0 is active, 1 is owned, 2-5 are flashes
qpic_t	*sb_ammo[4];
qpic_t	*sb_sigil[4];
qpic_t	*sb_armor[3];
qpic_t	*sb_items[32];

qpic_t	*sb_faces[7][2];		// 0 is gibbed, 1 is dead, 2-6 are alive
							// 0 is static, 1 is temporary animation
qpic_t	*sb_face_invis;
qpic_t	*sb_face_quad;
qpic_t	*sb_face_invuln;
qpic_t	*sb_face_invis_invuln;

qboolean	sb_showscores;

cvar_t *hud_background, *hud_inv_background;
cvar_t *hud_health, *hud_health_posx, *hud_health_posy, *hud_health_icon;
cvar_t *hud_armor, *hud_armor_posx, *hud_armor_posy, *hud_armor_icon;
cvar_t *hud_ammo, *hud_ammo_posx, *hud_ammo_posy, *hud_ammo_icon;
cvar_t *hud_inv, *hud_inv_posx, *hud_inv_posy, *hud_inv_numbers;

cvar_t *hud_health_align;
cvar_t *hud_armor_align;
cvar_t *hud_ammo_align;
cvar_t *hud_inv_align;


void Sbar_MiniDeathmatchOverlay (void);
void Sbar_DeathmatchOverlay (void);
void M_DrawPic (int x, int y, qpic_t *pic);
/*
=============
Draw_SubPic
=============
*/
void Draw_SubPic(int x, int y, qpic_t *pic, int srcx, int srcy, int width, int height)
{
	byte			*dest, *source;
	unsigned short	*pusdest;
	int				v, u;

	if ((x < 0) ||
		(x + width > vid.width) ||
		(y < 0) ||
		(y + height > vid.height))
	{
		Sys_Error ("Draw_SubPic: bad coordinates");
	}

	source = pic->data + srcy * pic->width + srcx;

	
	{
		dest = vid.buffer + y * vid.rowbytes + x;

		for (v=0 ; v<height ; v++)
		{
			Q_memcpy (dest, source, width);
			dest += vid.rowbytes;
			source += pic->width;
		}
	}

}

void Draw_SubPic_Scaled(int x, int y, qpic_t *pic, int srcx, int srcy, int width, int height)
{
	byte			*dest, *source;
	unsigned short	*pusdest;
	int				v, u, s;
	float vmax, umax;

	if ((x < 0) ||
		(x + width > vid.width) ||
		(y < 0) ||
		(y + height > vid.height))
	{
		Sys_Error ("Draw_SubPic: bad coordinates");
	}

	source = pic->data + srcy * pic->width + srcx;

	umax = pic->width * vid.width / (float)vid.vconwidth;
	vmax = pic->height * vid.height / (float)vid.vconheight;
	

	{
		byte *dest = vid.buffer + (y*vid.height/vid.vconheight) * vid.rowbytes + (x*vid.width/vid.vconwidth);
		for (v = 0; v < vmax; v++)
		{
			for (u = 0; u < umax; u++) 
			{
				s = u * vid.vconwidth / vid.width + (v * vid.vconheight/vid.height) * pic->width;
				dest[u] = source[s];
			}
			dest += vid.rowbytes;
		}
	}

}

/*
=============
Sbar_DrawSubPic
=============
JACK: Draws a portion of the picture in the status bar.
*/

void Sbar_DrawSubPic(int x, int y, qpic_t *pic, int srcx, int srcy, int width, int height) 
{
	if (sb_scaled){
	//Draw_SubPic_Scaled (x, y+(vid.vconheight-SBAR_HEIGHT), pic, srcx, srcy, vid.vconwidth, vid.vconheight);
	}else
	Draw_SubPic (x, y+(vid.height-SBAR_HEIGHT), pic, srcx, srcy, width, height);

}

/*
===============
Sbar_ShowScores

Tab key down
===============
*/
void Sbar_ShowScores (void)
{
	if (sb_showscores)
		return;
	sb_showscores = true;
	sb_updates = 0;
}

/*
===============
Sbar_DontShowScores

Tab key up
===============
*/
void Sbar_DontShowScores (void)
{
	sb_showscores = false;
	sb_updates = 0;
}


/*
===============
Sbar_Changed
===============
*/
void Sbar_Changed (void)
{
	sb_updates = 0;	// update next frame
}


/*
===============
Sbar_Init
===============
*/

#ifndef PROTO 

void Sbar_Init (void)
{
	int		i;

	for (i=0 ; i<10 ; i++)
	{
		sb_nums[0][i] = Draw_PicFromWad (va("num_%i",i));
	//
		sb_nums[1][i] = Draw_PicFromWad (va("anum_%i",i));
	}

	sb_nums[0][10] = Draw_PicFromWad ("num_minus");
	sb_nums[1][10] = Draw_PicFromWad ("anum_minus");

	sb_colon = Draw_PicFromWad ("num_colon");
	sb_slash = Draw_PicFromWad ("num_slash");

	sb_weapons[0][0] = Draw_PicFromWad ("inv_shotgun");
	sb_weapons[0][1] = Draw_PicFromWad ("inv_sshotgun");
	sb_weapons[0][2] = Draw_PicFromWad ("inv_nailgun");
	sb_weapons[0][3] = Draw_PicFromWad ("inv_snailgun");
	sb_weapons[0][4] = Draw_PicFromWad ("inv_rlaunch");
	sb_weapons[0][5] = Draw_PicFromWad ("inv_srlaunch");
	sb_weapons[0][6] = Draw_PicFromWad ("inv_lightng");

	sb_weapons[1][0] = Draw_PicFromWad ("inv2_shotgun");
	sb_weapons[1][1] = Draw_PicFromWad ("inv2_sshotgun");
	sb_weapons[1][2] = Draw_PicFromWad ("inv2_nailgun");
	sb_weapons[1][3] = Draw_PicFromWad ("inv2_snailgun");
	sb_weapons[1][4] = Draw_PicFromWad ("inv2_rlaunch");
	sb_weapons[1][5] = Draw_PicFromWad ("inv2_srlaunch");
	sb_weapons[1][6] = Draw_PicFromWad ("inv2_lightng");

	for (i=0 ; i<5 ; i++)
	{
		sb_weapons[2+i][0] = Draw_PicFromWad (va("inva%i_shotgun",i+1));
		sb_weapons[2+i][1] = Draw_PicFromWad (va("inva%i_sshotgun",i+1));
		sb_weapons[2+i][2] = Draw_PicFromWad (va("inva%i_nailgun",i+1));
		sb_weapons[2+i][3] = Draw_PicFromWad (va("inva%i_snailgun",i+1));
		sb_weapons[2+i][4] = Draw_PicFromWad (va("inva%i_rlaunch",i+1));
		sb_weapons[2+i][5] = Draw_PicFromWad (va("inva%i_srlaunch",i+1));
		sb_weapons[2+i][6] = Draw_PicFromWad (va("inva%i_lightng",i+1));
	}

	sb_ammo[0] = Draw_PicFromWad ("sb_shells");
	sb_ammo[1] = Draw_PicFromWad ("sb_nails");
	sb_ammo[2] = Draw_PicFromWad ("sb_rocket");
	sb_ammo[3] = Draw_PicFromWad ("sb_cells");

	sb_armor[0] = Draw_PicFromWad ("sb_armor1");
	sb_armor[1] = Draw_PicFromWad ("sb_armor2");
	sb_armor[2] = Draw_PicFromWad ("sb_armor3");

	sb_items[0] = Draw_PicFromWad ("sb_key1");
	sb_items[1] = Draw_PicFromWad ("sb_key2");
	sb_items[2] = Draw_PicFromWad ("sb_invis");
	sb_items[3] = Draw_PicFromWad ("sb_invuln");
	sb_items[4] = Draw_PicFromWad ("sb_suit");
	sb_items[5] = Draw_PicFromWad ("sb_quad");

	sb_sigil[0] = Draw_PicFromWad ("sb_sigil1");
	sb_sigil[1] = Draw_PicFromWad ("sb_sigil2");
	sb_sigil[2] = Draw_PicFromWad ("sb_sigil3");
	sb_sigil[3] = Draw_PicFromWad ("sb_sigil4");

	sb_faces[4][0] = Draw_PicFromWad ("face1");
	sb_faces[4][1] = Draw_PicFromWad ("face_p1");
	sb_faces[3][0] = Draw_PicFromWad ("face2");
	sb_faces[3][1] = Draw_PicFromWad ("face_p2");
	sb_faces[2][0] = Draw_PicFromWad ("face3");
	sb_faces[2][1] = Draw_PicFromWad ("face_p3");
	sb_faces[1][0] = Draw_PicFromWad ("face4");
	sb_faces[1][1] = Draw_PicFromWad ("face_p4");
	sb_faces[0][0] = Draw_PicFromWad ("face5");
	sb_faces[0][1] = Draw_PicFromWad ("face_p5");

	sb_face_invis = Draw_PicFromWad ("face_invis");
	sb_face_invuln = Draw_PicFromWad ("face_invul2");
	sb_face_invis_invuln = Draw_PicFromWad ("face_inv2");
	sb_face_quad = Draw_PicFromWad ("face_quad");

	Cmd_AddCommand ("+showscores", Sbar_ShowScores);
	Cmd_AddCommand ("-showscores", Sbar_DontShowScores);

	sb_sbar = Draw_PicFromWad ("sbar");
	sb_ibar = Draw_PicFromWad ("ibar");
	sb_scorebar = Draw_PicFromWad ("scorebar");

	hud_health = Cvar_Get ("hud_health", "1", CVAR_ARCHIVE | CVAR_ORIGINAL);
	hud_health_icon = Cvar_Get ("hud_health_icon", "1", CVAR_ARCHIVE | CVAR_ORIGINAL);
	hud_health_posx = Cvar_Get ("hud_health_posx", "112", CVAR_ARCHIVE | CVAR_ORIGINAL);
	hud_health_posy = Cvar_Get ("hud_health_posy", "0", CVAR_ARCHIVE | CVAR_ORIGINAL);
	hud_health_align = Cvar_Get ("hud_health_align", "0", CVAR_ARCHIVE | CVAR_ORIGINAL);

	hud_armor = Cvar_Get ("hud_armor", "1", CVAR_ARCHIVE | CVAR_ORIGINAL);
	hud_armor_icon = Cvar_Get ("hud_armor_icon", "1", CVAR_ARCHIVE | CVAR_ORIGINAL);
	hud_armor_posx = Cvar_Get ("hud_armor_posx", "0", CVAR_ARCHIVE | CVAR_ORIGINAL);
	hud_armor_posy = Cvar_Get ("hud_armor_posy", "0", CVAR_ARCHIVE | CVAR_ORIGINAL);
	hud_armor_align = Cvar_Get ("hud_armor_align", "0", CVAR_ARCHIVE | CVAR_ORIGINAL);

	hud_ammo = Cvar_Get ("hud_ammo", "1", CVAR_ARCHIVE | CVAR_ORIGINAL);
	hud_ammo_icon = Cvar_Get ("hud_ammo_icon", "1", CVAR_ARCHIVE | CVAR_ORIGINAL);
	hud_ammo_posx = Cvar_Get ("hud_ammo_posx", "224", CVAR_ARCHIVE | CVAR_ORIGINAL);
	hud_ammo_posy = Cvar_Get ("hud_ammo_posy", "0", CVAR_ARCHIVE | CVAR_ORIGINAL);
	hud_ammo_align = Cvar_Get ("hud_ammo_align", "0", CVAR_ARCHIVE | CVAR_ORIGINAL);

	hud_inv = Cvar_Get ("hud_inv", "1", CVAR_ARCHIVE | CVAR_ORIGINAL);
	hud_inv_numbers = Cvar_Get ("hud_inv_numbers", "1", CVAR_ARCHIVE | CVAR_ORIGINAL);
	hud_inv_posx = Cvar_Get ("hud_inv_posx", "0", CVAR_ARCHIVE | CVAR_ORIGINAL);
	hud_inv_posy = Cvar_Get ("hud_inv_posy", "-24", CVAR_ARCHIVE | CVAR_ORIGINAL);
	hud_inv_align = Cvar_Get ("hud_inv_align", "0", CVAR_ARCHIVE | CVAR_ORIGINAL);

	hud_background = Cvar_Get ("hud_background", "0", CVAR_ARCHIVE | CVAR_ORIGINAL);
	hud_inv_background = Cvar_Get ("hud_inv_background", "0", CVAR_ARCHIVE | CVAR_ORIGINAL);
}

#endif

//=============================================================================

// drawing routines are relative to the status bar location
/*
=============
Sbar_DrawPic
=============
*/
void Sbar_DrawPic (int x, int y, qpic_t *pic, int align)
{
	if (cl.gametype != GAME_DEATHMATCH)
		x += ((vid.vconwidth - 320)>>1);

	if(align)
		y += SBAR_HEIGHT;
	else
		y += (vid.vconheight-SBAR_HEIGHT);

	if (sb_scaled)
		Draw_Pic_Scaled (x, y,  pic);
	else
		Draw_Pic (x, y, pic);
	
}

void Sbar_DrawPicHalf (int x, int y, qpic_t *pic, int align)
{
	if (cl.gametype != GAME_DEATHMATCH)
		x += ((vid.vconwidth - 320)>>1);

	if(align)
		y += SBAR_HEIGHT;
	else
		y += (vid.vconheight-SBAR_HEIGHT);

	if (sb_scaled)
		Draw_Pic_Scaled_Two (x, y, pic);
	else
		Draw_Pic (x, y, pic);
}


/*x
================
Sbar_DrawCharacter

Draws one solid graphics character
================
*/
void Sbar_DrawCharacter (int x, int y, int num, int align)
{

	if (cl.gametype != GAME_DEATHMATCH)
		x += ((vid.vconwidth - 320)>>1);

	if(align)
		y += SBAR_HEIGHT;
	else
		y += (vid.vconheight-SBAR_HEIGHT);


	if (sb_scaled)
		Draw_Character_Scaled ( x /*+ ((vid.vconwidth - 320)>>1) */ + 4 , y, num);
	else
		Draw_Character ( x /*+ ((vid.width - 320)>>1) */ + 4 , y, num);
}

/*
================
Sbar_DrawString
================
*/
void Sbar_DrawString (int x, int y, char *str)
{
	if (sb_scaled){
	if (cl.gametype == GAME_DEATHMATCH)
		Draw_String_Scaled (x /*+ ((vid.vconwidth - 320)>>1)*/, y+ vid.vconheight-SBAR_HEIGHT, str);
	else
		Draw_String_Scaled (x + ((vid.vconwidth - 320)>>1), y+ vid.vconheight-SBAR_HEIGHT, str);
	}
	else
	{
	if (cl.gametype == GAME_DEATHMATCH)
		Draw_String (x /*+ ((vid.width - 320)>>1)*/, y+ vid.height-SBAR_HEIGHT, str);
	else
		Draw_String (x + ((vid.width - 320)>>1), y+ vid.height-SBAR_HEIGHT, str);
	}
}

/*
=============
Sbar_itoa
=============
*/
int Sbar_itoa (int num, char *buf)
{
	char	*str;
	int		pow10;
	int		dig;

	str = buf;

	if (num < 0)
	{
		*str++ = '-';
		num = -num;
	}

	for (pow10 = 10 ; num >= pow10 ; pow10 *= 10)
	;

	do
	{
		pow10 /= 10;
		dig = num/pow10;
		*str++ = '0'+dig;
		num -= dig*pow10;
	} while (pow10 != 1);

	*str = 0;

	return str-buf;
}


/*
=============
Sbar_DrawNum
=============
*/


void Sbar_DrawNum (int x, int y, int num, int digits, int color, int align)
{
	char			str[12];
	char			*ptr;
	int				l, frame;

	l = Sbar_itoa (num, str);
	ptr = str;
	if (l > digits)
		ptr += (l-digits);
	if (l < digits)
		x += (digits-l)*24;

	while (*ptr)
	{
		if (*ptr == '-')
			frame = STAT_MINUS;
		else
			frame = *ptr -'0';

	//Sbar_DrawPic (x,y,sb_nums[color][frame]);
		Sbar_DrawPic (x,y,sb_nums[color][frame], align);
		x += 24;
		ptr++;
	}
}


//=============================================================================

int		fragsort[MAX_SCOREBOARD];

char	scoreboardtext[MAX_SCOREBOARD][20];
int		scoreboardtop[MAX_SCOREBOARD];
int		scoreboardbottom[MAX_SCOREBOARD];
int		scoreboardcount[MAX_SCOREBOARD];
int		scoreboardlines;

/*
===============
Sbar_SortFrags
===============
*/
void Sbar_SortFrags (void)
{
	int		i, j, k;

// sort by frags
	scoreboardlines = 0;
	for (i=0 ; i<cl.maxclients ; i++)
	{
		if (cl.scores[i].name[0])
		{
			fragsort[scoreboardlines] = i;
			scoreboardlines++;
		}
	}

	for (i=0 ; i<scoreboardlines ; i++)
		for (j=0 ; j<scoreboardlines-1-i ; j++)
			if (cl.scores[fragsort[j]].frags < cl.scores[fragsort[j+1]].frags)
			{
				k = fragsort[j];
				fragsort[j] = fragsort[j+1];
				fragsort[j+1] = k;
			}
}

int	Sbar_ColorForMap (int m)
{
	return m < 128 ? m + 8 : m + 8;
}

/*
===============
Sbar_UpdateScoreboard
===============
*/
void Sbar_UpdateScoreboard (void)
{
	int		i, k;
	int		top, bottom;
	scoreboard_t	*s;

	Sbar_SortFrags ();

// draw the text
	memset (scoreboardtext, 0, sizeof(scoreboardtext));

	for (i=0 ; i<scoreboardlines; i++)
	{
		k = fragsort[i];
		s = &cl.scores[k];
		sprintf (&scoreboardtext[i][1], "%3i %s", s->frags, s->name);

		top = s->colors & 0xf0;
		bottom = (s->colors & 15) <<4;
		scoreboardtop[i] = Sbar_ColorForMap (top);
		scoreboardbottom[i] = Sbar_ColorForMap (bottom);
	}
}



/*
===============
Sbar_SoloScoreboard
===============
*/
void Sbar_SoloScoreboard (void)
{
	char	str[80];
	int		minutes, seconds, tens, units;
	int		l;

	sprintf (str,"Monsters:%3i /%3i", cl.stats[STAT_MONSTERS], cl.stats[STAT_TOTALMONSTERS]);
	Sbar_DrawString (8, 4, str);

	sprintf (str,"Secrets :%3i /%3i", cl.stats[STAT_SECRETS], cl.stats[STAT_TOTALSECRETS]);
	Sbar_DrawString (8, 12, str);

// time
	minutes = cl.time / 60;
	seconds = cl.time - 60*minutes;
	tens = seconds / 10;
	units = seconds - 10*tens;
	sprintf (str,"Time :%3i:%i%i", minutes, tens, units);
	Sbar_DrawString (184, 4, str);

// draw level name
	l = strlen (cl.levelname);
	Sbar_DrawString (232 - l*4, 12, cl.levelname);
}

/*
===============
Sbar_DrawScoreboard
===============
*/

extern int deathcam_yesiamdead;
extern float deathcam_whenidied;

void Sbar_DrawScoreboard (void)
{
	Sbar_SoloScoreboard ();
	if (cl.gametype == GAME_DEATHMATCH){
		if (deathcam_yesiamdead && ((sv.time - deathcam_whenidied) > 2))
		Sbar_DeathmatchOverlay ();
		else if (!deathcam_yesiamdead)
			Sbar_DeathmatchOverlay ();

	}
}

//=============================================================================

/*
===============
Sbar_DrawInventory
===============
*/
void Sbar_DrawInventory (void)
{
	int	i;
	char	num[6];
	float	time;
	int	flashon;

	if(!hud_inv->value)
		return;

	// Background
	if(hud_inv_background->value)
		Sbar_DrawPic (0, hud_inv_posy->value, sb_ibar, (int)hud_inv_align->value);

// weapons
	for (i=0 ; i<7 ; i++)
	{
		if (cl.items & (IT_SHOTGUN<<i) )
		{
			time = cl.item_gettime[i];
			flashon = (int)((cl.time - time)*10);
			if (flashon >= 10)
			{
				if ( cl.stats[STAT_ACTIVEWEAPON] == (IT_SHOTGUN<<i)  )
					flashon = 1;
				else
					flashon = 0;
			}
			else
				flashon = (flashon%5) + 2;

			Sbar_DrawPic (i*24, hud_inv_posy->value+8, sb_weapons[flashon][i], (int)hud_inv_align->value);

			if (flashon > 1)
				sb_updates = 0;		// force update to remove flash
		}
	}

	if(hud_inv_numbers->value)
	for (i=0 ; i<4 ; i++)
	{
		sprintf (num, "%3i",cl.stats[STAT_SHELLS+i] );
		if (num[0] != ' ')
			Sbar_DrawCharacter ( (6*i+1)*8 - 2, hud_inv_posy->value, 18 + num[0] - '0', (int)hud_inv_align->value);
		if (num[1] != ' ')
			Sbar_DrawCharacter ( (6*i+2)*8 - 2, hud_inv_posy->value, 18 + num[1] - '0', (int)hud_inv_align->value);
		if (num[2] != ' ')
			Sbar_DrawCharacter ( (6*i+3)*8 - 2, hud_inv_posy->value, 18 + num[2] - '0', (int)hud_inv_align->value);
		
	}

	flashon = 0;
	// items
	for (i=0 ; i<6 ; i++)
		if (cl.items & (1<<(17+i)))
		{
			time = cl.item_gettime[17+i];
			if (time && time > cl.time - 2 && flashon )
			{	// flash frame
				sb_updates = 0;
			}
			else
			{
				//MED 01/04/97 changed keys
				if ((i>1))
				{
					Sbar_DrawPic (192 + i*16, hud_inv_posy->value + 4, sb_items[i], (int)hud_inv_align->value);
				}
			}
			if (time && time > cl.time - 2)
				sb_updates = 0;
		}
	// sigils
	for (i=0 ; i<4 ; i++)
	{
		if (cl.items & (1<<(28+i)))
		{
			time = cl.item_gettime[28+i];
			if (time &&	time > cl.time - 2 && flashon )
			{	// flash frame
				sb_updates = 0;
			}
			else
				Sbar_DrawPic (320-32 + i*8, hud_inv_posy->value + 4, sb_sigil[i], (int)hud_inv_align->value);
			if (time &&	time > cl.time - 2)
				sb_updates = 0;
		}
	}
	
}

//=============================================================================

/*
===============
Sbar_DrawFrags
===============
*/
void Sbar_DrawFrags (void)
{
	int				i, k, l;
	int				top, bottom;
	int				x, y, f;
	int				xofs;
	char			num[12];
	scoreboard_t	*s;

	Sbar_SortFrags ();

// draw the text
	l = scoreboardlines <= 4 ? scoreboardlines : 4;
	if (sb_scaled){
	x = 23;
	if (cl.gametype == GAME_DEATHMATCH)
		xofs = 0;
	else
		xofs = (vid.vconwidth - 320)>>1;
	y = vid.vconheight - SBAR_HEIGHT - 23;
	}
	else
	{
	x = 23;
	if (cl.gametype == GAME_DEATHMATCH)
		xofs = 0;
	else
		xofs = (vid.width - 320)>>1;
	y = vid.height - SBAR_HEIGHT - 23;


	}

	for (i=0 ; i<l ; i++)
	{
		k = fragsort[i];
		s = &cl.scores[k];
		if (!s->name[0])
			continue;

	// draw background
		top = s->colors & 0xf0;
		bottom = (s->colors & 15)<<4;
		top = Sbar_ColorForMap (top);
		bottom = Sbar_ColorForMap (bottom);

		if (sb_scaled){
		Draw_Fill_Scaled (xofs + x*8 + 10, y, 28, 4, top);
		Draw_Fill_Scaled (xofs + x*8 + 10, y+4, 28, 3, bottom);
		}
		else
		{
		Draw_Fill (xofs + x*8 + 10, y, 28, 4, top);
		Draw_Fill (xofs + x*8 + 10, y+4, 28, 3, bottom);
		}
	// draw number
		f = s->frags;
		sprintf (num, "%3i",f);

		Sbar_DrawCharacter ( (x+1)*8 , -24, num[0], 0);
		Sbar_DrawCharacter ( (x+2)*8 , -24, num[1], 0);
		Sbar_DrawCharacter ( (x+3)*8 , -24, num[2], 0);

		if (k == cl.viewentity - 1)
		{
			Sbar_DrawCharacter (x*8+2, -24, 16, 0);
			Sbar_DrawCharacter ( (x+4)*8-4, -24, 17, 0);
		}
		x+=4;
	}
}

//=============================================================================


/*
===============
Sbar_DrawFace
===============
*/
void Sbar_DrawFace (void)
{
	int		f, anim;


	if ( (cl.items & (IT_INVISIBILITY | IT_INVULNERABILITY) )
	== (IT_INVISIBILITY | IT_INVULNERABILITY) )
	{
		Sbar_DrawPic (hud_health_posx->value, hud_health_posy->value, sb_face_invis_invuln, (int)hud_health_align->value);
		return;
	}
	if (cl.items & IT_QUAD)
	{
		Sbar_DrawPic (hud_health_posx->value, hud_health_posy->value, sb_face_quad, (int)hud_health_align->value);
		return;
	}
	if (cl.items & IT_INVISIBILITY)
	{
		Sbar_DrawPic (hud_health_posx->value, hud_health_posy->value, sb_face_invis, (int)hud_health_align->value);
		return;
	}
	if (cl.items & IT_INVULNERABILITY)
	{
		Sbar_DrawPic (hud_health_posx->value, hud_health_posy->value, sb_face_invuln, (int)hud_health_align->value);
		return;
	}

	if (cl.stats[STAT_HEALTH] >= 100)
		f = 4;
	else
		f = cl.stats[STAT_HEALTH] / 20;

	if (cl.time <= cl.faceanimtime)
	{
		anim = 1;
		sb_updates = 0;		// make sure the anim gets drawn over
	}
	else
		anim = 0;
	Sbar_DrawPic (hud_health_posx->value, hud_health_posy->value, sb_faces[f][anim], (int)hud_health_align->value);
}

/*
=============
Sbar_DrawNormal
=============
*/
void Sbar_DrawNormal (void)
{
	// Background
	if(hud_background->value)
		Sbar_DrawPic (0, 0, sb_sbar, 0);

// armor
	if(hud_armor->value && cl.stats[STAT_ARMOR])
	if (cl.items & IT_INVULNERABILITY)
	{
		Sbar_DrawNum (hud_armor_posx->value + 24, hud_armor_posy->value, 666, 3, 1, (int)hud_armor_align->value);

		if(hud_armor_icon->value)
			Sbar_DrawPic (hud_armor_posx->value, hud_armor_posy->value, draw_disc, (int)hud_armor_align->value);
	}
	else
	{
		Sbar_DrawNum (hud_armor_posx->value + 24, hud_armor_posy->value, cl.stats[STAT_ARMOR], 3
		, cl.stats[STAT_ARMOR] <= 25, (int)hud_armor_align->value);

		if(hud_armor_icon->value)
		if (cl.items & IT_ARMOR3)
			Sbar_DrawPic (hud_armor_posx->value, hud_armor_posy->value, sb_armor[2], (int)hud_armor_align->value);
		else if (cl.items & IT_ARMOR2)
			Sbar_DrawPic (hud_armor_posx->value, hud_armor_posy->value, sb_armor[1], (int)hud_armor_align->value);
		else if (cl.items & IT_ARMOR1)
			Sbar_DrawPic (hud_armor_posx->value, hud_armor_posy->value, sb_armor[0], (int)hud_armor_align->value);
	}
	

	if(hud_health->value)
	{
		if(hud_health_icon->value)
			Sbar_DrawFace ();

		Sbar_DrawNum (hud_health_posx->value + 24, hud_health_posy->value, cl.stats[STAT_HEALTH], 3, cl.stats[STAT_HEALTH] <= 25, (int)hud_health_align->value);
	}

// ammo icon
	if(hud_ammo->value && cl.stats[STAT_AMMO])
	{
		if(hud_ammo_icon->value)
		if (cl.items & IT_SHELLS)
			Sbar_DrawPic (hud_ammo_posx->value, hud_ammo_posy->value, sb_ammo[0], (int)hud_ammo_align->value);
		else if (cl.items & IT_NAILS)
			Sbar_DrawPic (hud_ammo_posx->value, hud_ammo_posy->value, sb_ammo[1], (int)hud_ammo_align->value);
		else if (cl.items & IT_ROCKETS)
			Sbar_DrawPic (hud_ammo_posx->value, hud_ammo_posy->value, sb_ammo[2], (int)hud_ammo_align->value);
		else if (cl.items & IT_CELLS)
			Sbar_DrawPic (hud_ammo_posx->value, hud_ammo_posy->value, sb_ammo[3], (int)hud_ammo_align->value);

		Sbar_DrawNum (hud_ammo_posx->value + 24, hud_ammo_posy->value, cl.stats[STAT_AMMO], 3
		, cl.stats[STAT_AMMO] <= 10, (int)hud_ammo_align->value);
	}
}

void Sbar_Draw (void)
{
	if(!(cl_sbar->value))
		return;

	qboolean headsup;
	char st[512];

	headsup = !(cl_sbar->value);

	if (scr_con_current == vid.height)
		return;		// console is full screen


	if ((sb_updates >= vid.numpages) && !headsup)
		return;

	scr_copyeverything = 1;

	sb_updates++;

	Sbar_DrawInventory ();

	if (cl.maxclients != 1)			
		Sbar_DrawFrags ();
		

	 if (sb_showscores || cl.stats[STAT_HEALTH] <= 0)
			Sbar_SoloScoreboard ();
		else
			Sbar_DrawNormal ();
	

	if (sb_showscores || cl.stats[STAT_HEALTH] <= 0)
	{
		Sbar_DrawPic (0, 0, sb_scorebar, 0);
		Sbar_DrawScoreboard ();
		sb_updates = 0;
	}

	if (vid.width > 320) {
		if (cl.gametype == GAME_DEATHMATCH)
			Sbar_MiniDeathmatchOverlay ();
	}
}

//=============================================================================

/*
==================
Sbar_IntermissionNumber

==================
*/
void Sbar_IntermissionNumber (int x, int y, int num, int digits, int color)
{
	char			str[12];
	char			*ptr;
	int				l, frame;

	l = Sbar_itoa (num, str);
	ptr = str;
	if (l > digits)
		ptr += (l-digits);
	if (l < digits)
		x += (digits-l)*24;

	while (*ptr)
	{
		if (*ptr == '-')
			frame = STAT_MINUS;
		else
			frame = *ptr -'0';

		Draw_Pic (x,y,sb_nums[color][frame]);
		x += 24;
		ptr++;
	}
}


/*
==================
Sbar_IntermissionNumber

==================
*/
void Sbar_IntermissionNumber_Scaled (int x, int y, int num, int digits, int color)
{
	char			str[12];
	char			*ptr;
	int				l, frame;

	l = Sbar_itoa (num, str);
	ptr = str;
	if (l > digits)
		ptr += (l-digits);
	if (l < digits)
		x += (digits-l)*24;

	while (*ptr)
	{
		if (*ptr == '-')
			frame = STAT_MINUS;
		else
			frame = *ptr -'0';

		Draw_Pic_Scaled (x,y,sb_nums[color][frame]);
		x += 24;
		ptr++;
	}
}
/*
==================
Sbar_DeathmatchOverlay

==================
*/
void Sbar_DeathmatchOverlay (void)
{
	qpic_t			*pic;
	int				i, k, l;
	int				top, bottom;
	int				x, y, f;
	char			num[12];
	scoreboard_t	*s;

	scr_copyeverything = 1;
	scr_fullupdate = 0;

	pic = Draw_CachePic ("gfx/ranking.lmp");
	M_DrawPic ((320-pic->width)/2, 8, pic);

// scores
	Sbar_SortFrags ();

// draw the text
	l = scoreboardlines;

	if (sb_scaled) x = 80 + ((vid.vconwidth - 320)>>1);
	else x = 80 + ((vid.width - 320)>>1);

	y = 40;
	for (i=0 ; i<l ; i++)
	{
		k = fragsort[i];
		s = &cl.scores[k];
		if (!s->name[0])
			continue;

	// draw background
		top = s->colors & 0xf0;
		bottom = (s->colors & 15)<<4;
		top = Sbar_ColorForMap (top);
		bottom = Sbar_ColorForMap (bottom);

		if (sb_scaled){
		Draw_Fill_Scaled ( x, y, 40, 4, top);
		Draw_Fill_Scaled ( x, y+4, 40, 4, bottom);
		} else {
		Draw_Fill ( x, y, 40, 4, top);
		Draw_Fill ( x, y+4, 40, 4, bottom);
		}

		// draw number
		f = s->frags;
		sprintf (num, "%3i",f);

		if (sb_scaled){
		Draw_Character_Scaled ( x+8 , y, num[0]);
		Draw_Character_Scaled ( x+16 , y, num[1]);
		Draw_Character_Scaled ( x+24 , y, num[2]);

		if (k == cl.viewentity - 1)
			Draw_Character_Scaled ( x - 8, y, 12);
		}
		else
		{
		Draw_Character ( x+8 , y, num[0]);
		Draw_Character ( x+16 , y, num[1]);
		Draw_Character ( x+24 , y, num[2]);

		if (k == cl.viewentity - 1)
			Draw_Character ( x - 8, y, 12);
		}

	// draw name
		if (sb_scaled)
		Draw_String_Scaled (x+64, y, s->name);
		else Draw_String (x+64, y, s->name);
		y += 10;
	}
}

/*
==================
Sbar_DeathmatchOverlay

==================
*/
void Sbar_MiniDeathmatchOverlay (void)
{
//	qpic_t			*pic;	// 2000-07-30 DJGPP compiler warning fix by Norberto Alfredo Bensa
	int				i, k;	//, l	// 2001-12-10 Reduced compiler warnings by Jeff Ford
	int				top, bottom;
	int				x, y, f;
	char			num[12];
	scoreboard_t	*s;
	int				numlines;

	if (sb_scaled){
		if (vid.vconwidth < 512)
		return;
	}
	else
	{
		if (vid.width < 512)
		return;
	}

	scr_copyeverything = 1;
	scr_fullupdate = 0;

// scores
	Sbar_SortFrags ();

// draw the text
	y = vid.vconheight - 24;
	numlines = 24/8;
	if (numlines < 3)
		return;

	//find us
	for (i = 0; i < scoreboardlines; i++)
		if (fragsort[i] == cl.viewentity - 1)
			break;

    if (i == scoreboardlines) // we're not there
		i = 0;
    else // figure out start
		i = i - numlines/2;

    if (i > scoreboardlines - numlines)
		i = scoreboardlines - numlines;
    if (i < 0)
		i = 0;

	x = 324;
	for (/* */; i < scoreboardlines && y < vid.vconheight - 8 ; i++)
	{
		k = fragsort[i];
		s = &cl.scores[k];
		if (!s->name[0])
			continue;

	// draw background
		top = s->colors & 0xf0;
		bottom = (s->colors & 15)<<4;
		top = Sbar_ColorForMap (top);
		bottom = Sbar_ColorForMap (bottom);


		if(sb_scaled){
		Draw_Fill_Scaled ( x, y+1, 40, 3, top);
		Draw_Fill_Scaled ( x, y+4, 40, 4, bottom);

	// draw number
		f = s->frags;
		sprintf (num, "%3i",f);
		Draw_Character_Scaled ( x+8 , y, num[0]);
		Draw_Character_Scaled ( x+16 , y, num[1]);
		Draw_Character_Scaled ( x+24 , y, num[2]);

		if (k == cl.viewentity - 1) {
			Draw_Character_Scaled ( x, y, 16);
			Draw_Character_Scaled ( x + 32, y, 17);
		}

	// draw name
		Draw_String_Scaled (x+48, y, s->name);
		}
		else
		{

		Draw_Fill ( x, y+1, 40, 3, top);
		Draw_Fill ( x, y+4, 40, 4, bottom);

	// draw number
		f = s->frags;
		sprintf (num, "%3i",f);

		Draw_Character ( x+8 , y, num[0]);
		Draw_Character ( x+16 , y, num[1]);
		Draw_Character ( x+24 , y, num[2]);

		if (k == cl.viewentity - 1) {
			Draw_Character ( x, y, 16);
			Draw_Character ( x + 32, y, 17);
		}

	// draw name
		Draw_String (x+48, y, s->name);


		}
		y += 8;
	}
}

/*
==================
Sbar_IntermissionOverlay

==================
*/

void Sbar_IntermissionOverlay (void)
{
	qpic_t	*pic;
	int		dig;
	int		num;

	scr_copyeverything = 1;
	scr_fullupdate = 0;

	if (cl.gametype == GAME_DEATHMATCH)
	{
		
		Sbar_DeathmatchOverlay ();
		return;
	}

	if (sb_scaled){
	pic = Draw_CachePic ("gfx/complete.lmp");
	Draw_Pic_Scaled (64, 24, pic);

	pic = Draw_CachePic ("gfx/inter.lmp");
	Draw_Pic_Scaled (0, 56, pic);

// time
	dig = cl.completed_time/60;
	Sbar_IntermissionNumber_Scaled (160, 64, dig, 3, 0);
	num = cl.completed_time - dig*60;
	Draw_Pic_Scaled (234,64,sb_colon);
	Draw_Pic_Scaled (246,64,sb_nums[0][num/10]);
	Draw_Pic_Scaled (266,64,sb_nums[0][num%10]);

	Sbar_IntermissionNumber_Scaled (160, 104, cl.stats[STAT_SECRETS], 3, 0);
	Draw_Pic_Scaled (232,104,sb_slash);
	Sbar_IntermissionNumber_Scaled (240, 104, cl.stats[STAT_TOTALSECRETS], 3, 0);

	Sbar_IntermissionNumber_Scaled (160, 144, cl.stats[STAT_MONSTERS], 3, 0);
	Draw_Pic_Scaled (232,144,sb_slash);
	Sbar_IntermissionNumber_Scaled (240, 144, cl.stats[STAT_TOTALMONSTERS], 3, 0);
	}
	else
	{
	pic = Draw_CachePic ("gfx/complete.lmp");
	Draw_Pic (64, 24, pic);

	pic = Draw_CachePic ("gfx/inter.lmp");
	Draw_Pic (0, 56, pic);

// time
	dig = cl.completed_time/60;
	Sbar_IntermissionNumber (160, 64, dig, 3, 0);
	num = cl.completed_time - dig*60;
	Draw_Pic (234,64,sb_colon);
	Draw_Pic (246,64,sb_nums[0][num/10]);
	Draw_Pic (266,64,sb_nums[0][num%10]);

	Sbar_IntermissionNumber (160, 104, cl.stats[STAT_SECRETS], 3, 0);
	Draw_Pic (232,104,sb_slash);
	Sbar_IntermissionNumber (240, 104, cl.stats[STAT_TOTALSECRETS], 3, 0);

	Sbar_IntermissionNumber (160, 144, cl.stats[STAT_MONSTERS], 3, 0);
	Draw_Pic (232,144,sb_slash);
	Sbar_IntermissionNumber (240, 144, cl.stats[STAT_TOTALMONSTERS], 3, 0);

	}
}


/*
==================
Sbar_FinaleOverlay

==================
*/
void Sbar_FinaleOverlay (void)
{
	qpic_t	*pic;

	scr_copyeverything = 1;

	pic = Draw_CachePic ("gfx/finale.lmp");

	if(sb_scaled)
	Draw_Pic_Scaled ( (vid.vconwidth-pic->width)/2, 16, pic);
	else
		Draw_Pic ( (vid.width-pic->width)/2, 16, pic);
}
