/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// globaldef.h -- primary header for client
	
#define	QUAKE_GAME			// as opposed to utilities

#ifdef _WIN32
	#include "mathlib.h"		// I make djgpp upset for some reason
#endif

#include "matrixlib.h"
#include "version.h"

#define ITSFIX
#define REALLYCRASHESITHINK
// -----------------------------------------
// Video Features
// -----------------------------------------

#define INTERPOL7		// other interpolation again (still buggy)*WIP*

// -----------------------------------------
// Audio Features
// -----------------------------------------

//#define DUMB				// Use DUMB Module playback library *WIP*
#ifndef linux
	#define ASS_MIDI			// Use Apogee Sound System for MIDI playback only*WIP*
#endif
int inthedos;
//#define NOASM

#define EFFINGMOUSE			// disables "Enhance Pointer Precision" 
					// a definite MUST for debugging

// -----------------------------------------
// Gameplay Features
// -----------------------------------------

#define	GLOBOT				// tomaz globots
//#define VMTOC				// viewmodelforclient changes
#define SONOFABITS			// Bits workaround :(
//#define	BITSAGAIN		// qip way

// -----------------------------------------
// Versioning
// -----------------------------------------

int	protocol;		// OK.
int Nehahrademcompatibility; 	// LordHavoc: to allow playback of the early Nehahra movie segments
int dpprotocol;	

#define	DPPROTOCOLS	
#define	VERSION			0.5

#define	PROTOCOL_STOCK		0		// old 1.09
#define	PROTOCOL_QUAKEDP	1		// dp105
#define	PROTOCOL_TQ		2		// Tomaz

// -----------------------------------------
// Game Data
// -----------------------------------------

#define	GAMENAME	"data"

#include <math.h>
#include <string.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <setjmp.h>

//
// The precompiler definition "__i386__" comes from the DJGPP package and
// states that the code is compiled for an Intel x386 compatible machine.
// Windows compilers (definition "_WIN32") do not provide this DJGPP definition,
// but provide something similar called "_M_IX86". So when using a Windows
// compiler "__i386__" can be created when "_WIN32" and "_M_IX86" are defined.
//
// If you do not want to use assembler code, then set the definition "id386" to
// zero (0) in QUAKEASM.H for asm sources and in QUAKEDEF.H for C sources
//

#if defined(_WIN32) && !defined(WINDED)
	#if defined(_M_IX86)
		#define __i386__	1
	#endif

	void	VID_LockBuffer (void);
	void	VID_UnlockBuffer (void);
#else
	#define	VID_LockBuffer()
	#define	VID_UnlockBuffer()
#endif

#if defined __i386__ && !defined NOASM // && !defined __sun__
	#define id386		1
	#define id386poly 	0
	#define id386rgb	0
#else
	#define id386		0
	#define id386poly 	0
	#define id386rgb	0
#endif

#if id386
	#define UNALIGNED_OK	1	// set to 0 if unaligned accesses are not supported
#else
	#define UNALIGNED_OK	0
#endif

// !!! if this is changed, it must be changed in d_ifacea.h too !!!
#define CACHE_SIZE	32		// used to align key data structures

#define UNUSED(x)	(x = x)	// for pesky compiler / lint warnings

#ifdef BENCH
	#define	MINIMUM_MEMORY			0x0186A0 // 100kb hack test
#else
	#define	MINIMUM_MEMORY			0x550000
#endif

#define	MINIMUM_MEMORY_LEVELPAK	(MINIMUM_MEMORY + 0x100000)
#define MAX_NUM_ARGVS	50

#define	PITCH 	0 // up / down
#define	YAW 	1 // left / right
#define	ROLL	2 // fall over

#define	MAX_QPATH	64		// max length of a quake game pathname
#define	MAX_OSPATH	128		// max length of a filesystem pathname
#define	ON_EPSILON	0.1		// point on plane side epsilon

#ifdef QSB_NET
	#define	MAX_MSGLEN	65536		// max length of a reliable message
	#define	MAX_DATAGRAM	1400		// max length of unreliable message
#else
	#define	MAX_MSGLEN	8000		// max length of a reliable message
	#define	MAX_DATAGRAM	1024		// max length of unreliable message
#endif




//
// per-level limits
//
// 2001-09-20 Configurable entity limits by Maddes  start
//#define	MAX_EDICTS		600			// FIXME: ouch! ouch! ouch!


#ifdef QSB_NET
#define	MAX_MODELS		4096		// these are sent over the net as bytes
#define	MAX_SOUNDS		4096		// so they cannot be blindly increased

#define	MIN_EDICTS		600		// must be the original default value: 600
#define	MAX_EDICTS		8192		// Maximum without network changes, as entity number is send as signed short (2 bytes)

#define	MIN_TEMP_ENTITIES	1024		// lightning bolts, etc
#define	MIN_STATIC_ENTITIES	1024		// torches, etc
// 2001-09-20 Configurable entity limits by Maddes  end
#define	MAX_LIGHTSTYLES	64


#else
#define	MAX_MODELS		256		// these are sent over the net as bytes
#define	MAX_SOUNDS		256		// so they cannot be blindly increased

#define	MIN_EDICTS		600		// must be the original default value: 600
#define	MAX_EDICTS		0x7FFF		// Maximum without network changes, as entity number is send as signed short (2 bytes)

#define	MIN_TEMP_ENTITIES	64		// lightning bolts, etc
#define	MIN_STATIC_ENTITIES	128		// torches, etc
// 2001-09-20 Configurable entity limits by Maddes  end
#define	MAX_LIGHTSTYLES	64

#endif

#define	SAVEGAME_COMMENT_LENGTH	39

#define	MAX_STYLESTRING	64

//
// stats are integers communicated to the client by the server
//
#define	MAX_CL_STATS		32
#define	STAT_HEALTH		0
#define	STAT_FRAGS		1
#define	STAT_WEAPON		2
#define	STAT_AMMO		3
#define	STAT_ARMOR		4
#define	STAT_WEAPONFRAME	5
#define	STAT_SHELLS		6
#define	STAT_NAILS		7
#define	STAT_ROCKETS		8
#define	STAT_CELLS		9
#define	STAT_ACTIVEWEAPON	10
#define	STAT_TOTALSECRETS	11
#define	STAT_TOTALMONSTERS	12
#define	STAT_SECRETS		13		// bumped on client side by svc_foundsecret
#define	STAT_MONSTERS		14		// bumped by svc_killedmonster

// stock defines

#define	IT_SHOTGUN		1
#define	IT_SUPER_SHOTGUN	2
#define	IT_NAILGUN		4
#define	IT_SUPER_NAILGUN	8
#define	IT_GRENADE_LAUNCHER	16
#define	IT_ROCKET_LAUNCHER	32
#define	IT_LIGHTNING		64
#define IT_SUPER_LIGHTNING	128
#define IT_SHELLS		256
#define IT_NAILS		512
#define IT_ROCKETS		1024
#define IT_CELLS		2048
#define IT_AXE			4096
#define IT_ARMOR1		8192
#define IT_ARMOR2		16384
#define IT_ARMOR3		32768
#define IT_SUPERHEALTH		65536
#define IT_KEY1			131072
#define IT_KEY2			262144
#define	IT_INVISIBILITY		524288
#define	IT_INVULNERABILITY	1048576
#define	IT_SUIT			2097152
#define	IT_QUAD			4194304
#define IT_SIGIL1		(1<<28)
#define IT_SIGIL2		(1<<29)
#define IT_SIGIL3		(1<<30)
#define IT_SIGIL4		(1<<31)

#define	MAX_SCOREBOARD		16
#define	MAX_SCOREBOARDNAME	32

#define	SOUND_CHANNELS		8

#include "common.h"
#include "bspfile.h"
#include "vid.h"
#include "sys.h"
#include "zone.h"
#include "mathlib.h"
#include "matrixlib.h"

typedef struct
{
	vec3_t	origin;
	vec3_t	angles;
	int		modelindex;
	int		frame;
	int		colormap;
	int		skin;
	int		effects;
} entity_state_t;

#include "wad.h"
#include "draw.h"
#include "cvar.h"
#include "screen.h"
#include "net.h"
#include "protocol.h"
#include "cmd.h"
#include "sbar.h"
#include "sound.h"
#include "render.h"
#include "client.h"
#include "progs.h"
#include "server.h"

#include "nvs_common.h"		// 2000-04-30 NVS COMMON by Maddes

#include "model.h"
#include "d_iface.h"
#include "model_common.h"	// 2001-12-28 Merged model functions by Maddes

#include "input.h"
#include "world.h"
#include "keys.h"
#include "console.h"
#include "view.h"
#include "menu.h"
#include "crc.h"
#include "cdaudio.h"
#include "tracker.h"

//=============================================================================

// the host system specifies the base of the directory tree, the
// command line parms passed to the program, and the amount of memory
// available for the program to use

typedef struct
{
	char	*basedir;
	char	*cachedir;		// for development over ISDN lines
	int		argc;
	char	**argv;
	void	*membase;
	int		memsize;
} quakeparms_t;


//=============================================================================



extern qboolean noclip_anglehack;


//
// host
//
extern	quakeparms_t host_parms;

extern	cvar_t		*sys_ticrate;
extern	cvar_t		*sys_nostdout;
extern	cvar_t		*developer;

extern	qboolean	host_initialized;		// true if into command execution

extern	double		host_frametime;
extern	byte		*host_basepal;
extern	byte		*host_colormap;
extern	byte		*host_colormap_red;
extern	byte		*host_colormap_green;
extern	byte		*host_colormap_blue;
extern	byte		*host_colormap_buffer;
extern	byte		*host_fogmap;		// leilei - fog
extern	byte		*host_egamap;
extern	byte		*glcolormap;
extern	byte		*host_noopal;		// for translations
extern	int		host_fullbrights;   // for preserving fullbrights in color operations

extern	int		host_framecount;	// incremented every frame, never reset
extern	double		realtime;			// not bounded in any way, changed at
										// start of every frame, never reset

void Host_ClearMemory (void);
void Host_ServerFrame (void);
void Host_InitCommands (void);
void Host_Init (quakeparms_t *parms);
void Host_Shutdown(void);
void Host_Error (char *error, ...);
void Host_EndGame (char *message, ...);
void Host_Frame (float time);
void Host_Quit_f (void);
void Host_ClientCommands (char *fmt, ...);
void Host_ShutdownServer (qboolean crash);
void Host_Version_f (void);				// 2000-04-30 NVS HANDSHAKE SRV<->CL/QC<->CL by Maddes

extern qboolean		msg_suppress_1;		// suppresses resolution and cache size console output
extern int		current_skill;		// skill level for currently loaded level (in case
										//  the user changes the cvar while the level is
										//  running, this reflects the level actually in use)

extern qboolean		isDedicated;
extern int		minimum_memory;

//
// chase
//
extern	cvar_t	*chase_active;

void Chase_Init (void);
void Chase_Reset (void);
void Chase_Update (void);

extern qboolean		nouse;	// 1999-10-29 +USE fix by Maddes

#define PROGHEADER090_CRC	26940
#define PROGHEADER091_CRC	26940
#define PROGHEADER092_CRC	26940	

// 1999-10-28 Compatibilty check by Maddes  start
#define PROGHEADER101_CRC	5927
extern qboolean		keep_compatibility;
// 1999-10-28 Compatibilty check by Maddes  end

// 2000-07-30 DJGPP compiler warning fix by Norberto Alfredo Bensa  start
qboolean SV_RecursiveHullCheck (hull_t *hull, int num, float p1f, float p2f, vec3_t p1, vec3_t p2, trace_t *trace);
dfunction_t *ED_FindFunction (char *name);	// 2000-01-09 QCExec by FrikaC/Maddes
// 2000-07-30 DJGPP compiler warning fix by Norberto Alfredo Bensa  end

// 2001-10-20 TIMESCALE extension by Tomaz/Maddes  start
extern	double	host_cpu_frametime;
extern	double	host_org_frametime;

extern	cvar_t	*host_timescale;
// 2001-10-20 TIMESCALE extension by Tomaz/Maddes  end

typedef unsigned char UBYTE;  // a one-byte int 
typedef unsigned short UWORD; // a two-byte int 
//typedef unsigned int ULONG;   // a four-byte int (assumes int 4 bytes) 
typedef unsigned int OOLONG;   // a four-byte int (assumes int 4 bytes) 
