/*
Copyright (C) 2015 Marco "eukara" Hladik

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

/* 
 * XMMS Music playback for the engine
 *
 * Unlike the XMMS-CTL extension, this one actually talks to the player with libxmms, sadly
 * this will only work with 32bit installations of XMMS which are rather uncommon these days.
 */

#include "globaldef.h"
#include <xmms/xmmsctrl.h>

unsigned int xmms_session = 0;
static int xmms_loop = 1;
static int xmms_track;

static qboolean	xmms_playing = false;
static qboolean	xmms_wasplaying = false;
static float xmms_volume = 1.0f;

void Tracker_Play(byte track, qboolean looping)
{
	xmms_remote_set_playlist_pos(xmms_session, (int) track);
	xmms_remote_play(xmms_session);
	xmms_loop = looping == true ? 0 : 1;
	xmms_track = (int) track;
	xmms_playing = true;
	xmms_volume = bgmvolume->value;
}

void Tracker_Stop(void)
{
	xmms_remote_stop(xmms_session);
	xmms_wasplaying = false;
	xmms_playing = false;
}


void Tracker_Pause(void)
{
	xmms_remote_pause(xmms_session);
	xmms_wasplaying = xmms_playing;
	xmms_playing = false;
}


void Tracker_Resume(void)
{
	if (!xmms_wasplaying)
		return;
	xmms_remote_play(xmms_session);
	xmms_playing = false;

}

// DO NOTHING
void Tracker_Update(void)
{
	if(xmms_loop)
	if(xmms_remote_get_playlist_pos(xmms_session) != xmms_track)
		xmms_remote_set_playlist_pos(xmms_session, xmms_track);

	if(xmms_volume != bgmvolume->value)
	{
		xmms_volume = bgmvolume->value;
		xmms_remote_set_main_volume(xmms_session, (int)(xmms_volume * 100));
	}
}

static void XMMS_f (void)
{
	char	*command;
	int i;

	if (Cmd_Argc() < 2)
		return;

	command = Cmd_Argv (1);

	if (Q_strcasecmp(command, "play") == 0)
		Tracker_Play((byte)Q_atoi(Cmd_Argv (2)), false);
	else if (Q_strcasecmp(command, "stop") == 0)
		Tracker_Stop();
	else if (Q_strcasecmp(command, "pause") == 0)
		Tracker_Pause();
	else if (Q_strcasecmp(command, "resume") == 0)
		Tracker_Resume();
	else if (Q_strcasecmp(command, "quit") == 0)
		xmms_remote_quit(xmms_session);
	else if (Q_strcasecmp(command, "info") == 0)
	{
		Con_Printf("%u tracks\n", xmms_remote_get_playlist_length(xmms_session));
		if (xmms_playing)
			Con_Printf("Currently %s track %u (%s)\n", xmms_loop ? "looping" : "playing", xmms_track, xmms_remote_get_playlist_title(xmms_session, xmms_track));
		else if (xmms_wasplaying)
			Con_Printf("Paused %s track %u (%s)\n", xmms_loop ? "looping" : "playing", xmms_track, xmms_remote_get_playlist_title(xmms_session, xmms_track));
		Con_Printf("Volume is %f\n", xmms_remote_get_main_volume(xmms_session));
		return;
	}
	else if(Q_strcasecmp(command, "tracks") == 0)
	for(i = 0; i < xmms_remote_get_playlist_length(xmms_session); i++)
		Con_Printf("#%i %s\n", i, xmms_remote_get_playlist_title(xmms_session, i));
}

int Tracker_Init(void)
{
	Cmd_AddCommand ("xmms", XMMS_f);
	return 1;
}


void Tracker_Shutdown(void)
{
	Tracker_Stop();
}
